package sun.font;

import java.awt.geom.NoninvertibleTransformException;
import java.util.concurrent.ConcurrentHashMap;
import java.lang.ref.SoftReference;
import java.awt.geom.AffineTransform;
import java.awt.font.FontRenderContext;
import java.awt.Font;
import java.awt.geom.Point2D;
import java.util.ArrayList;

public final class GlyphLayout
{
    private GlyphLayout.GVData _gvdata;
    private static volatile GlyphLayout cache;
    private GlyphLayout.LayoutEngineFactory _lef;
    private TextRecord _textRecord;
    private ScriptRun _scriptRuns;
    private FontRunIterator _fontRuns;
    private int _ercount;
    private ArrayList _erecords;
    private Point2D.Float _pt;
    private FontStrikeDesc _sd;
    private float[] _mat;
    private int _typo_flags;
    private int _offset;
    
    public static GlyphLayout get(GlyphLayout.LayoutEngineFactory instance) {
        if (instance == null) {
            instance = SunLayoutEngine.instance();
        }
        Object cache = null;
        synchronized (GlyphLayout.class) {
            if (GlyphLayout.cache != null) {
                cache = GlyphLayout.cache;
                GlyphLayout.cache = null;
            }
        }
        if (cache == null) {
            cache = new GlyphLayout();
        }
        cache._lef = instance;
        return cache;
    }
    
    public static void done(GlyphLayout cache) {
        cache._lef = null;
        GlyphLayout.cache = cache;
    }
    
    public StandardGlyphVector layout(Font obj, FontRenderContext fontRenderContext, char[] array, int offset, int n, int n2, StandardGlyphVector standardGlyphVector) {
        if (array == null || offset < 0 || n < 0 || n > array.length - offset) {
            throw new IllegalArgumentException();
        }
        this.init(n);
        if (obj.hasLayoutAttributes()) {
            AttributeValues values = ((AttributeMap)obj.getAttributes()).getValues();
            if (values.getKerning() != 0) {
                this._typo_flags |= 0x1;
            }
            if (values.getLigatures() != 0) {
                this._typo_flags |= 0x2;
            }
        }
        this._offset = offset;
        GlyphLayout.SDCache value = GlyphLayout.SDCache.get(obj, fontRenderContext);
        this._mat[0] = (float)value.gtx.getScaleX();
        this._mat[1] = (float)value.gtx.getShearY();
        this._mat[2] = (float)value.gtx.getShearX();
        this._mat[3] = (float)value.gtx.getScaleY();
        this._pt.setLocation(value.delta);
        int n3 = offset + n;
        int n4 = 0;
        int length = array.length;
        if (n2 != 0) {
            if ((n2 & 0x1) != 0x0) {
                this._typo_flags |= Integer.MIN_VALUE;
            }
            if ((n2 & 0x2) != 0x0) {
                n4 = offset;
            }
            if ((n2 & 0x4) != 0x0) {
                length = n3;
            }
        }
        int n5 = -1;
        Font2D font2D = FontUtilities.getFont2D(obj);
        if ((font2D instanceof FontSubstitution)) {
            font2D = ((FontSubstitution)font2D).getCompositeFont2D();
        }
        this._textRecord.init(array, offset, n3, n4, length);
        int n6 = offset;
        if ((font2D instanceof CompositeFont)) {
            this._scriptRuns.init(array, offset, n);
            this._fontRuns.init((CompositeFont)font2D, array, offset, n3);
            while (this._scriptRuns.next()) {
                int pos;
                for (int scriptLimit2 = this._scriptRuns.getScriptLimit(), scriptCode = this._scriptRuns.getScriptCode(); this._fontRuns.next(scriptCode, scriptLimit2); n6 = pos) {
                    PhysicalFont physicalFont = this._fontRuns.getFont();
                    if ((physicalFont instanceof NativeFont)) {
                        physicalFont = ((NativeFont)physicalFont).getDelegateFont();
                    }
                    int glyphMask = this._fontRuns.getGlyphMask();
                    pos = this._fontRuns.getPos();
                    this.nextEngineRecord(n6, pos, scriptCode, n5, physicalFont, glyphMask);
                }
            }
        }
        else {
            this._scriptRuns.init(array, offset, n);
            while (this._scriptRuns.next()) {
                int scriptLimit = this._scriptRuns.getScriptLimit();
                this.nextEngineRecord(n6, scriptLimit, this._scriptRuns.getScriptCode(), n5, font2D, 0);
                n6 = scriptLimit;
            }
        }
        int i = 0;
        int ercount = this._ercount;
        int n7 = 1;
        if (this._typo_flags < 0) {
            i = ercount - 1;
            ercount = -1;
            n7 = -1;
        }
        this._sd = value.sd;
        while (i != ercount) {
            GlyphLayout.EngineRecord engineRecord = (GlyphLayout.EngineRecord)this._erecords.get(i);
            Label_0559: {
                try {
                    engineRecord.layout();
                }
                catch (IndexOutOfBoundsException ex) {
                    if (this._gvdata._count >= 0) {
                        this._gvdata.grow();
                    }
                    continue;
                }
            }
            if (this._gvdata._count < 0) {
                break;
            }
            i += n7;
        }
        Object glyphVector;
        if (this._gvdata._count < 0) {
            glyphVector = new StandardGlyphVector(obj, array, offset, n, fontRenderContext);
            if (FontUtilities.debugFonts()) {
                FontUtilities.getLogger().warning("OpenType layout failed on font: " + obj);
            }
        }
        else {
            glyphVector = this._gvdata.createGlyphVector(obj, fontRenderContext, standardGlyphVector);
        }
        return glyphVector;
    }
    
    private GlyphLayout() {
        this._gvdata = new GlyphLayout.GVData();
        this._textRecord = new TextRecord();
        this._scriptRuns = new ScriptRun();
        this._fontRuns = new FontRunIterator();
        this._erecords = new ArrayList(10);
        this._pt = new Point2D.Float();
        this._sd = new FontStrikeDesc();
        this._mat = new float[4];
    }
    
    private void init(int n) {
        this._typo_flags = 0;
        this._ercount = 0;
        this._gvdata.init(n);
    }
    
    private void nextEngineRecord(int n, int n2, int n3, int n4, Font2D font2D, int n5) {
        GlyphLayout.EngineRecord e;
        if (this._ercount == this._erecords.size()) {
            e = new GlyphLayout.EngineRecord();
            this._erecords.add(e);
        }
        else {
            e = (GlyphLayout.EngineRecord)this._erecords.get(this._ercount);
        }
        e.init(n, n2, font2D, n3, n4, n5);
        ++this._ercount;
    }
    
    public static final class LayoutEngineKey
    {
        private Font2D font;
        private int script;
        private int lang;
        
        LayoutEngineKey() {
        }
        
        LayoutEngineKey(Font2D font2D, int n, int n2) {
            this.init(font2D, n, n2);
        }
        
        void init(Font2D font, int script, int lang) {
            this.font = font;
            this.script = script;
            this.lang = lang;
        }
        
        GlyphLayout.LayoutEngineKey copy() {
            return new GlyphLayout.LayoutEngineKey(this.font, this.script, this.lang);
        }
        
        Font2D font() {
            return this.font;
        }
        
        int script() {
            return this.script;
        }
        
        int lang() {
            return this.lang;
        }
        
        @Override
        public boolean equals(Object o) {
            if (this == o) {
                return true;
            }
            if (o == null) {
                return false;
            }
            try {
                GlyphLayout.LayoutEngineKey layoutEngineKey = (GlyphLayout.LayoutEngineKey)o;
                return this.script == layoutEngineKey.script && this.lang == layoutEngineKey.lang && this.font.equals(layoutEngineKey.font);
            }
            catch (ClassCastException ex) {
                return false;
            }
        }
        
        @Override
        public int hashCode() {
            return this.script ^ this.lang ^ this.font.hashCode();
        }
    }
    
    private static final class SDCache
    {
        public Font key_font;
        public FontRenderContext key_frc;
        public AffineTransform dtx;
        public AffineTransform invdtx;
        public AffineTransform gtx;
        public Point2D.Float delta;
        public FontStrikeDesc sd;
        private static final Point2D.Float ZERO_DELTA;
        private static SoftReference<ConcurrentHashMap<GlyphLayout.SDCache.SDKey, GlyphLayout.SDCache>> cacheRef;
        
        private SDCache(Font key_font, FontRenderContext key_frc) {
            this.key_font = key_font;
            this.key_frc = key_frc;
            (this.dtx = key_frc.getTransform()).setTransform(this.dtx.getScaleX(), this.dtx.getShearY(), this.dtx.getShearX(), this.dtx.getScaleY(), 0.0, 0.0);
            if (!this.dtx.isIdentity()) {
                try {
                    this.invdtx = this.dtx.createInverse();
                }
                catch (NoninvertibleTransformException cause) {
                    throw new InternalError(cause);
                }
            }
            float size2D = key_font.getSize2D();
            if (key_font.isTransformed()) {
                (this.gtx = key_font.getTransform()).scale((double)size2D, (double)size2D);
                this.delta = new Point2D.Float((float)this.gtx.getTranslateX(), (float)this.gtx.getTranslateY());
                this.gtx.setTransform(this.gtx.getScaleX(), this.gtx.getShearY(), this.gtx.getShearX(), this.gtx.getScaleY(), 0.0, 0.0);
                this.gtx.preConcatenate(this.dtx);
            }
            else {
                this.delta = GlyphLayout.SDCache.ZERO_DELTA;
                (this.gtx = new AffineTransform(this.dtx)).scale((double)size2D, (double)size2D);
            }
            this.sd = new FontStrikeDesc(this.dtx, this.gtx, key_font.getStyle(), FontStrikeDesc.getAAHintIntVal(key_frc.getAntiAliasingHint(), FontUtilities.getFont2D(key_font), (int)Math.abs(size2D)), FontStrikeDesc.getFMHintIntVal(key_frc.getFractionalMetricsHint()));
        }
        
        public static GlyphLayout.SDCache get(Font font, FontRenderContext fontRenderContext) {
            if (fontRenderContext.isTransformed()) {
                AffineTransform transform = fontRenderContext.getTransform();
                if (transform.getTranslateX() != 0.0 || transform.getTranslateY() != 0.0) {
                    fontRenderContext = new FontRenderContext(new AffineTransform(transform.getScaleX(), transform.getShearY(), transform.getShearX(), transform.getScaleY(), 0.0, 0.0), fontRenderContext.getAntiAliasingHint(), fontRenderContext.getFractionalMetricsHint());
                }
            }
            GlyphLayout.SDCache.SDKey o = new GlyphLayout.SDCache.SDKey(font, fontRenderContext);
            Object referent = null;
            Object value = null;
            if (GlyphLayout.SDCache.cacheRef != null) {
                referent = (ConcurrentHashMap<GlyphLayout.SDCache.SDKey, GlyphLayout.SDCache>)GlyphLayout.SDCache.cacheRef.get();
                if (referent != null) {
                    value = (Object)referent.get(o);
                }
            }
            if (value == null) {
                value = new GlyphLayout.SDCache(font, fontRenderContext);
                if (referent == null) {
                    referent = new ConcurrentHashMap(10);
                    GlyphLayout.SDCache.cacheRef = new SoftReference(referent);
                }
                else if (referent.size() >= 512) {
                    referent.clear();
                }
                referent.put(o, value);
                return value;
            }
            else {
                return value;
            }
        }
        
        static {
            ZERO_DELTA = new Point2D.Float();
        }
        
        private static final class SDKey
        {
            private final Font font;
            private final FontRenderContext frc;
            private final int hash;
            
            SDKey(Font font, FontRenderContext frc) {
                this.font = font;
                this.frc = frc;
                this.hash = (font.hashCode() ^ frc.hashCode());
            }
            
            @Override
            public int hashCode() {
                return this.hash;
            }
            
            @Override
            public boolean equals(Object o) {
                try {
                    GlyphLayout.SDCache.SDKey sdKey = (GlyphLayout.SDCache.SDKey)o;
                    return this.hash == sdKey.hash && this.font.equals(sdKey.font) && this.frc.equals(sdKey.frc);
                }
                catch (ClassCastException ex) {
                    return false;
                }
            }
        }
    }
    
    public static final class GVData
    {
        public int _count;
        public int _flags;
        public int[] _glyphs;
        public float[] _positions;
        public int[] _indices;
        private static final int UNINITIALIZED_FLAGS = -1;
        
        public void init(int n) {
            this._count = 0;
            this._flags = -1;
            if (this._glyphs == null || this._glyphs.length < n) {
                if (n < 20) {
                    n = 20;
                }
                this._glyphs = new int[n];
                this._positions = new float[n * 2 + 2];
                this._indices = new int[n];
            }
        }
        
        public void grow() {
            this.grow(this._glyphs.length / 4);
        }
        
        public void grow(int n) {
            int n2 = this._glyphs.length + n;
            int[] glyphs = new int[n2];
            System.arraycopy(this._glyphs, 0, glyphs, 0, this._count);
            this._glyphs = glyphs;
            float[] positions = new float[n2 * 2 + 2];
            System.arraycopy(this._positions, 0, positions, 0, this._count * 2 + 2);
            this._positions = positions;
            int[] indices = new int[n2];
            System.arraycopy(this._indices, 0, indices, 0, this._count);
            this._indices = indices;
        }
        
        public void adjustPositions(AffineTransform affineTransform) {
            affineTransform.transform(this._positions, 0, this._positions, 0, this._count);
        }
        
        public StandardGlyphVector createGlyphVector(Font font, FontRenderContext fontRenderContext, StandardGlyphVector standardGlyphVector) {
            if (this._flags == -1) {
                this._flags = 0;
                if (this._count > 1) {
                    int n = 1;
                    int n2 = 1;
                    int n4;
                    for (int count = this._count, n3 = 0; n3 < this._count && (n != 0 || n2 != 0); n = (n != 0 && n4 == n3), n2 = (n2 != 0 && n4 == --count), ++n3) {
                        n4 = this._indices[n3];
                    }
                    if (n2 != 0) {
                        this._flags |= 0x4;
                    }
                    if (n2 == 0 && n == 0) {
                        this._flags |= 0x8;
                    }
                }
                this._flags |= 0x2;
            }
            int[] array = new int[this._count];
            System.arraycopy(this._glyphs, 0, array, 0, this._count);
            float[] array2 = null;
            if ((this._flags & 0x2) != 0x0) {
                array2 = new float[this._count * 2 + 2];
                System.arraycopy(this._positions, 0, array2, 0, array2.length);
            }
            int[] array3 = null;
            if ((this._flags & 0x8) != 0x0) {
                array3 = new int[this._count];
                System.arraycopy(this._indices, 0, array3, 0, this._count);
            }
            if (standardGlyphVector == null) {
                standardGlyphVector = new StandardGlyphVector(font, fontRenderContext, array, array2, array3, this._flags);
                return standardGlyphVector;
            }
            standardGlyphVector.initGlyphVector(font, fontRenderContext, array, array2, array3, this._flags);
            return standardGlyphVector;
        }
    }
    
    private final class EngineRecord
    {
        private int start;
        private int limit;
        private int gmask;
        private int eflags;
        private GlyphLayout.LayoutEngineKey key;
        private GlyphLayout.LayoutEngine engine;
        
        EngineRecord() {
            this.key = new GlyphLayout.LayoutEngineKey();
        }
        
        void init(int i, int limit, Font2D font2D, int n, int n2, int gmask) {
            this.start = i;
            this.limit = limit;
            this.gmask = gmask;
            this.key.init(font2D, n, n2);
            this.eflags = 0;
            while (i < limit) {
                int codePoint = GlyphLayout.this._textRecord.text[i];
                if (Character.isHighSurrogate((char)codePoint) && i < limit - 1 && Character.isLowSurrogate(GlyphLayout.this._textRecord.text[i + 1])) {
                    codePoint = Character.toCodePoint((char)codePoint, GlyphLayout.this._textRecord.text[++i]);
                }
                int type = Character.getType(codePoint);
                if (type == 6 || type == 7 || type == 8) {
                    this.eflags = 4;
                    break;
                }
                ++i;
            }
            this.engine = GlyphLayout.this._lef.getEngine(this.key);
        }
        
        void layout() {
            GlyphLayout.this._textRecord.start = this.start;
            GlyphLayout.this._textRecord.limit = this.limit;
            this.engine.layout(GlyphLayout.this._sd, GlyphLayout.this._mat, this.gmask, this.start - GlyphLayout.this._offset, GlyphLayout.this._textRecord, GlyphLayout.this._typo_flags | this.eflags, GlyphLayout.this._pt, GlyphLayout.this._gvdata);
        }
    }
    
    public interface LayoutEngine
    {
        public abstract void layout(FontStrikeDesc p0, float[] p1, int p2, int p3, TextRecord p4, int p5, Point2D.Float p6, GlyphLayout.GVData p7);
    }
    
    public interface LayoutEngineFactory
    {
        public abstract GlyphLayout.LayoutEngine getEngine(Font2D p0, int p1, int p2);
        
        public abstract GlyphLayout.LayoutEngine getEngine(GlyphLayout.LayoutEngineKey p0);
    }
}
